
import torch
from torch_geometric.nn import MessagePassing, GCNConv
from torch_geometric.utils import add_self_loops, degree
import numpy as np
import torch.nn as nn
import torch.nn.functional as F

from settings import USE_SELF_EMBEDDING

class TransE(nn.Module):
  def __init__(self, entity_count, relation_count, device, norm=1, dim=100, margin=1.0):
    super(TransE, self).__init__()
    self.entity_count = entity_count
    self.relation_count = relation_count
    self.device = device
    self.norm = norm
    self.dim = dim
    self.entities_emb = self._init_enitity_emb()
    self.relations_emb = self._init_relation_emb()
    self.criterion = nn.MarginRankingLoss(margin=margin, reduction='none')

  def _init_enitity_emb(self):
    entities_emb = nn.Embedding(num_embeddings=self.entity_count + 1, embedding_dim=self.dim, padding_idx=self.entity_count)
    uniform_range = 6 / np.sqrt(self.dim)
    entities_emb.weight.data.uniform_(-uniform_range, uniform_range)
    return entities_emb

  def _init_relation_emb(self):
    relations_emb = nn.Embedding(num_embeddings=self.relation_count + 1, embedding_dim=self.dim, padding_idx=self.relation_count)
    uniform_range = 6 / np.sqrt(self.dim)
    relations_emb.weight.data.uniform_(-uniform_range, uniform_range)
    # -1 to avoid nan for OOV vector
    relations_emb.weight.data[:-1, :].div_(relations_emb.weight.data[:-1, :].norm(p=1, dim=1, keepdim=True))
    return relations_emb

  def forward(self, positive_triplets: torch.LongTensor, negative_triplets: torch.LongTensor):
    """
    Return model losses based on the input.
    :param positive_triplets: triplets of positives in Bx3 shape (B - batch, 3 - head, relation and tail)
    :param negative_triplets: triplets of negatives in Bx3 shape (B - batch, 3 - head, relation and tail)
    :return: tuple of the model loss, positive triplets loss component, negative triples loss component
    """
    # -1 to avoid nan for OOV vector
    self.entities_emb.weight.data[:-1, :].div_(self.entities_emb.weight.data[:-1, :].norm(p=2, dim=1, keepdim=True))

    assert positive_triplets.size()[1] == 3
    positive_distances = self._distance(positive_triplets)

    assert negative_triplets.size()[1] == 3
    negative_distances = self._distance(negative_triplets)

    return self.loss(positive_distances, negative_distances), positive_distances, negative_distances

  def predict(self, triplets: torch.LongTensor):
    """
    Calculated dissimilarity score for given triplets.
    :param triplets: triplets in Bx3 shape (B - batch, 3 - head, relation and tail)
    :return: dissimilarity score for given triplets
    """
    return self._distance(triplets)

  def loss(self, positive_distances, negative_distances):
    target = torch.tensor([-1], dtype=torch.long, device=self.device)
    return self.criterion(positive_distances, negative_distances, target)

  def _distance(self, triplets):
    """
    Triplets should have shape Bx3 where dim 3 are head id, relation id, tail id.
    """
    assert triplets.size()[1] == 3
    heads = triplets[:, 0]
    relations = triplets[:, 1]
    tails = triplets[:, 2]
    return (self.entities_emb(heads) + self.relations_emb(relations) - self.entities_emb(tails)).norm(p=self.norm, dim=1)


class BasePropagator(MessagePassing):
  def __init__(self, r_vector, is_pos, embedding_type):
    super().__init__(aggr='add')

    # Get relation embedding
    self.r_vector = r_vector

    # Get message direction
    self.is_pos = is_pos

    # Get embedding type
    self.embedding_type = embedding_type

  def forward(self, x, edge_index, data, subtract_1):
    # Do message-passing
    return self.propagate(edge_index, x=x, data=data, subtract_1=subtract_1)

  def message(self, x_j, data, subtract_1):
    # Calculate translated entities
    r_tmp = torch.reshape(self.r_vector(data.edge_type - subtract_1), (x_j.shape[0], -1))
    if self.is_pos:
      if self.embedding_type == 'ComplEx':
        halfway_point = int(int(x_j.shape[1]) / 2)
        #tmp1 = (x_j[:, 0:halfway_point] * r_tmp[:, 0:halfway_point]) - (x_j[:, halfway_point:int(x_j.shape[1])] * r_tmp[:, halfway_point:int(r_tmp.shape[1])])
        #tmp2 = (x_j[:, 0:halfway_point] * r_tmp[:, halfway_point:int(r_tmp.shape[1])]) + (x_j[:, halfway_point:int(x_j.shape[1])] * r_tmp[:, 0:halfway_point])
        tmp1 = ((x_j[:, 0:halfway_point]) * ((r_tmp[:, 0:halfway_point]) + (r_tmp[:, halfway_point:int(r_tmp.shape[1])]))) + ((x_j[:, halfway_point:int(x_j.shape[1])]) * ((r_tmp[:, 0:halfway_point]) - (r_tmp[:, halfway_point:int(r_tmp.shape[1])])))
        tmp2 = ((x_j[:, 0:halfway_point]) * ((r_tmp[:, 0:halfway_point]) - (r_tmp[:, halfway_point:int(r_tmp.shape[1])]))) - ((x_j[:, halfway_point:int(x_j.shape[1])]) * ((r_tmp[:, 0:halfway_point]) + (r_tmp[:, halfway_point:int(r_tmp.shape[1])])))
        tmp = torch.cat((tmp1, tmp2), 1)
      if self.embedding_type == 'TransE':
        tmp = x_j + r_tmp
      elif self.embedding_type == 'RotatE':
        m = nn.Tanh()
        tmp = x_j * (m(1000 * r_tmp))
      elif self.embedding_type == 'DistMult':
        tmp = x_j * r_tmp
      elif self.embedding_type == 'RESCAL':
        r_tmp = torch.unsqueeze(r_tmp, -1)
        r_tmp = torch.reshape(r_tmp, (x_j.shape[0], x_j.shape[1], x_j.shape[1]))
        x_j = torch.unsqueeze(x_j, 1)
        tmp = torch.squeeze(torch.bmm(x_j, r_tmp))
    else:
      if self.embedding_type == 'ComplEx':
        halfway_point = int(int(x_j.shape[1]) / 2)
        #tmp1 = (x_j[:, 0:halfway_point] * r_tmp[:, 0:halfway_point]) - (x_j[:, halfway_point:int(x_j.shape[1])] * r_tmp[:, halfway_point:int(r_tmp.shape[1])])
        #tmp2 = (x_j[:, 0:halfway_point] * r_tmp[:, halfway_point:int(r_tmp.shape[1])]) + (x_j[:, halfway_point:int(x_j.shape[1])] * r_tmp[:, 0:halfway_point])
        tmp1 = ((x_j[:, 0:halfway_point]) * ((r_tmp[:, 0:halfway_point]) - (r_tmp[:, halfway_point:int(r_tmp.shape[1])]))) + ((x_j[:, halfway_point:int(x_j.shape[1])]) * ((r_tmp[:, 0:halfway_point]) + (r_tmp[:, halfway_point:int(r_tmp.shape[1])])))
        tmp2 = ((x_j[:, 0:halfway_point]) * ((r_tmp[:, 0:halfway_point]) + (r_tmp[:, halfway_point:int(r_tmp.shape[1])]))) - ((x_j[:, halfway_point:int(x_j.shape[1])]) * ((r_tmp[:, 0:halfway_point]) - (r_tmp[:, halfway_point:int(r_tmp.shape[1])])))
        tmp = torch.cat((tmp1, tmp2), 1)
      if self.embedding_type == 'TransE':
        tmp = x_j - r_tmp
      elif self.embedding_type == 'RotatE':
        m = nn.Tanh()
        tmp = x_j / (m(1000 * r_tmp))
      elif self.embedding_type == 'DistMult':
        tmp = x_j * r_tmp
      elif self.embedding_type == 'RESCAL':
        r_tmp = torch.unsqueeze(r_tmp, -1)
        r_tmp = torch.reshape(r_tmp, (x_j.shape[0], x_j.shape[1], x_j.shape[1]))
        x_j = torch.unsqueeze(x_j, -1)
        tmp = torch.squeeze(torch.bmm(r_tmp, x_j))
    return tmp


class ShallowGNNLayer(torch.nn.Module):
  def __init__(self, in_channels, out_channels, r_vector, embedding_type, num_relations):
    super().__init__()

    # Learnable weights
    self.lin = torch.nn.Linear(in_channels, out_channels)

    # Get relation embedding
    self.r_vector = r_vector

    self.num_relations = num_relations

    # Get embedding type
    self.embedding_type = embedding_type

    # Set up forward and backward message propagators
    self.props = nn.ModuleList([BasePropagator(self.r_vector, True, self.embedding_type), BasePropagator(self.r_vector, False, self.embedding_type)])

  def forward(self, x, edge_index, edge_index_neg, data, subtract_1):
    # Copy features for backwards pass
    x_neg = torch.clone(x)

    # Do message-passing
    try:
      if USE_SELF_EMBEDDING:
        tempx = self.props[0](x, edge_index, data, subtract_1) + self.props[1](x_neg, edge_index_neg, data, subtract_1)
        r_tmp = torch.reshape(self.r_vector(torch.ones((x.shape[0],), dtype=torch.int) * self.num_relations), (x.shape[0], -1))
        if self.embedding_type == 'ComplEx':
          halfway_point = int(int(x.shape[1]) / 2)
          #tmp1 = (x[:, 0:halfway_point] * r_tmp[:, 0:halfway_point]) - (x[:, halfway_point:int(x.shape[1])] * r_tmp[:, halfway_point:int(r_tmp.shape[1])])
          #tmp2 = (x[:, 0:halfway_point] * r_tmp[:, halfway_point:int(r_tmp.shape[1])]) + (x[:, halfway_point:int(x.shape[1])] * r_tmp[:, 0:halfway_point])
          tmp1 = ((x[:, 0:halfway_point]) * ((r_tmp[:, 0:halfway_point]) + (r_tmp[:, halfway_point:int(r_tmp.shape[1])]))) + ((x[:, halfway_point:int(x.shape[1])]) * ((r_tmp[:, 0:halfway_point]) - (r_tmp[:, halfway_point:int(r_tmp.shape[1])])))
          tmp2 = ((x[:, 0:halfway_point]) * ((r_tmp[:, 0:halfway_point]) - (r_tmp[:, halfway_point:int(r_tmp.shape[1])]))) - ((x[:, halfway_point:int(x.shape[1])]) * ((r_tmp[:, 0:halfway_point]) + (r_tmp[:, halfway_point:int(r_tmp.shape[1])])))
          tmp = torch.cat((tmp1, tmp2), 1)
        if self.embedding_type == 'TransE':
          tmp = x + r_tmp
        elif self.embedding_type == 'RotatE':
          m = nn.Tanh()
          tmp = x * (m(1000 * r_tmp))
        elif self.embedding_type == 'DistMult':
          tmp = x * r_tmp
        elif self.embedding_type == 'RESCAL':
          r_tmp = torch.unsqueeze(r_tmp, -1)
          r_tmp = torch.reshape(r_tmp, (x.shape[0], x.shape[1], x.shape[1]))
          x_j = torch.unsqueeze(x, 1)
          tmp = torch.squeeze(torch.bmm(x, r_tmp))
        x = tmp + tempx
      else:
        x = x + self.props[0](x, edge_index, data, subtract_1) + self.props[1](x_neg, edge_index_neg, data, subtract_1)
    except IndexError:
      print('Cur X shape: ' + str(x.shape))
      print(self.r_vector)
      print(torch.max(data.edge_type - subtract_1))
      print(torch.min(data.edge_type - subtract_1))
      raise Exception('There was an index error here.')
    #x = x / 2.0
    x = x / (torch.reshape(degree(edge_index.flatten(), num_nodes=int(x.shape[0]), dtype=torch.float), (x.shape[0], -1)) + 1.0)

    # Apply learnable weights
    x = self.lin(x)
    return x


class ShallowGNN(torch.nn.Module):
  def __init__(self, in_channels, out_channels, num_layers, mid_dim, num_features, num_relations, embedding_type):
    super().__init__()

    # Get embedding type
    self.embedding_type = embedding_type
    self.num_relations = num_relations

    if USE_SELF_EMBEDDING:
      embedding_size = self.num_relations + 1
    else:
      embedding_size = self.num_relations + 1

    # Create relation embedding
    if self.embedding_type == 'ComplEx':
      self.r_vector = nn.Embedding(num_embeddings=embedding_size, embedding_dim=num_features*2)
    elif self.embedding_type == 'RESCAL':
      self.r_vector = nn.Embedding(num_embeddings=embedding_size, embedding_dim=num_features**2)
    else:
      self.r_vector = nn.Embedding(num_embeddings=embedding_size, embedding_dim=num_features)

    # Set up message-passing layers
    self.convs = []
    if self.embedding_type == 'ComplEx':
      self.convs.append(ShallowGNNLayer(in_channels*2, mid_dim*2, self.r_vector, self.embedding_type, self.num_relations))
    else:
      self.convs.append(ShallowGNNLayer(in_channels, mid_dim, self.r_vector, self.embedding_type, self.num_relations))
    for x in range(num_layers - 2):
      if self.embedding_type == 'ComplEx':
        self.convs.append(ShallowGNNLayer(mid_dim*2, mid_dim*2, self.r_vector, self.embedding_type, self.num_relations))
      else:
        self.convs.append(ShallowGNNLayer(mid_dim, mid_dim, self.r_vector, self.embedding_type, self.num_relations))
    if self.embedding_type == 'ComplEx':
      self.convs.append(ShallowGNNLayer(mid_dim*2, out_channels, self.r_vector, self.embedding_type, self.num_relations))
    else:
      self.convs.append(ShallowGNNLayer(mid_dim, out_channels, self.r_vector, self.embedding_type, self.num_relations))
    self.convs = nn.ModuleList(self.convs)

    # Learnable weights
    self.lin = torch.nn.Linear(2 + num_relations, 1)

  def encode(self, x, edge_index, data):
    # Calculate backwards edge_index
    if int(torch.min(data.edge_type - 1)) <= 0:
      subtract_1 = 0
    else:
      subtract_1 = 1
    edge_index_neg = torch.zeros(2, edge_index.shape[1], dtype=torch.long).index_copy_(0, torch.tensor([1, 0]), edge_index)

    # Do message-passing
    for conv in self.convs[0:-1]:
      x = conv(x, edge_index, edge_index_neg, data, subtract_1)
      x = F.relu(x)
    x = self.convs[-1](x, edge_index, edge_index_neg, data, subtract_1)
    return x

  def decode(self, z, pos_edge_index, neg_edge_index, pos_relations, neg_relations, num_relations): # only pos and neg edges
    edge_index = torch.cat([pos_edge_index, neg_edge_index], dim=-1) # concatenate pos and neg edges
    total_relations = torch.cat([pos_relations, neg_relations], dim=0)
    logits = (z[edge_index[0]] * z[edge_index[1]]).sum(dim=-1)  # dot product 
    logits = self.lin(torch.cat([torch.reshape(logits, (logits.shape[0], 1)), F.one_hot(total_relations, num_classes=num_relations)], dim=-1))
    return logits

  def decode_all(self, z):
    prob_adj = z @ z.t() # get adj NxN
    return (prob_adj > 0).nonzero(as_tuple=False).t() # get predicted edge_list 

  def get_self_embedding(self):
    return torch.reshape(self.r_vector(torch.tensor([self.num_relations])), (1, -1))


# Simple toy model for testing
class ToyNet(torch.nn.Module):
  def __init__(self):
    super(Net, self).__init__()
    self.conv1 = GCNConv(dataset.num_node_features, 128)
    self.conv2 = GCNConv(128, 64)

  def encode(self):
    x = self.conv1(data.x, data.train_pos_edge_index)
    x = x.relu()
    return self.conv2(x, data.train_pos_edge_index)

  def decode(self, z, pos_edge_index, neg_edge_index):
    edge_index = torch.cat([pos_edge_index, neg_edge_index], dim=-1)
    logits = (z[edge_index[0]] * z[edge_index[1]]).sum(dim=-1)
    return logits

  def decode_all(self, z):
    prob_adj = z @ z.t()
    return (prob_adj > 0).nonzero(as_tuple=False).t()

